#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <phidget22.h>

void closeAll(PhidgetRCServoHandle*, PhidgetVoltageRatioInputHandle*, PhidgetDigitalInputHandle);

#ifndef _WIN32
#include <unistd.h>
#else
#include <Windows.h>
#endif

static void CCONV ssleep(double);

#define NUM_SERVOS 12

typedef enum legServos {
	LFV = 0,
	LFH = 1,
	RFV = 2,
	RFH = 3,
	LMV = 4,
	LMH = 5,
	RMV = 6,
	RMH = 7,
	LRV = 8,
	LRH = 9,
	RRV = 10,
	RRH = 11
} LEG_SERVOS;

#define VMAX 180
#define HMAX 30

int legsDone[NUM_SERVOS] = { 0 };

typedef enum stepTypes {
	GROUP1 = 0,
	GROUP2 = 1,
	NEUTRAL = 2
} STEP_TYPES;

int stepTracker = NEUTRAL;
int crouched = 0;

int isLeftSide(int index) {
	if (index % 4 == 0 || (index + 1) % 4 == 0)
		return 1;
	return 0;
}

int isHorizontal(int index) {
	return index % 2;
}

void setLegPos(PhidgetRCServoHandle* servos, int index, double pos) {
	legsDone[index] = 0;

	if (isHorizontal(index)) {
		if (!isLeftSide(index))
			pos *= -1;
		PhidgetRCServo_setTargetPosition_async(servos[index], 90 + pos, NULL, NULL);
	} else {
		if (isLeftSide(index))
			pos = 180 - pos;
		PhidgetRCServo_setTargetPosition_async(servos[index], pos, NULL, NULL);
	}
}

int checkLegsDone() {
	for (int i = 0; i < NUM_SERVOS; i++)
		if (legsDone[i] == 0)
			return 0;

	return 1;
}

void waitLegsDone() {
	for (int i = 0; i < NUM_SERVOS; i += 2)
		while (legsDone[i] == 0)
			ssleep(0.02);
}

int tripod[2][2][3] = { { { LFV, RMV, LRV },{ LFH, RMH, LRH } },{ { RFV, LMV, RRV },{ RFH, LMH, RRH } } };

#define V 0
#define H 1

#define F 0
#define M 1
#define R 2

void walk(PhidgetRCServoHandle* servos, PhidgetVoltageRatioInputHandle* axes) {


	double turnOffset;
	double yAxis;
	double xAxis;
	double dir;

	PhidgetVoltageRatioInput_getVoltageRatio(axes[0], &xAxis);
	PhidgetVoltageRatioInput_getVoltageRatio(axes[1], &yAxis);

	turnOffset = xAxis * HMAX;

	dir = yAxis;

	if (!checkLegsDone())
		return;

	//If the joystick is centred, force input to zero 
	if (fabs(xAxis) < 0.1 && fabs(yAxis) < 0.1) {
		dir = 0;
		turnOffset = 0;
		//Set stepTracker to NEUTRAL after movement to allow legs to centre themselves first
	}

	//If stopped and there is new input, start with GROUP1
	if (stepTracker == NEUTRAL && dir != 0 && turnOffset != 0)
		stepTracker = GROUP1;

	//Perform the step
	if (stepTracker == GROUP1 || stepTracker == GROUP2) {
		if (stepTracker == GROUP2)
			turnOffset *= -1;

		//Start lifting legs
		setLegPos(servos, tripod[stepTracker][V][F], 70);
		setLegPos(servos, tripod[stepTracker][V][M], 70);
		setLegPos(servos, tripod[stepTracker][V][R], 70);

		waitLegsDone();

		//Move all legs to next positions, finish raising legs
		setLegPos(servos, tripod[stepTracker][V][F], VMAX);
		setLegPos(servos, tripod[stepTracker][V][M], VMAX);
		setLegPos(servos, tripod[stepTracker][V][R], VMAX);

		setLegPos(servos, tripod[stepTracker][H][F], HMAX * dir + turnOffset);
		setLegPos(servos, tripod[stepTracker][H][M], HMAX * dir - turnOffset / 4);
		setLegPos(servos, tripod[stepTracker][H][R], HMAX * dir + turnOffset);

		setLegPos(servos, tripod[!stepTracker][H][F], -HMAX * dir + turnOffset);
		setLegPos(servos, tripod[!stepTracker][H][M], -HMAX * dir - turnOffset);
		setLegPos(servos, tripod[!stepTracker][H][R], -HMAX * dir + turnOffset);

		waitLegsDone();

		//Put raised legs down
		setLegPos(servos, tripod[stepTracker][V][F], 0);
		setLegPos(servos, tripod[stepTracker][V][M], 0);
		setLegPos(servos, tripod[stepTracker][V][R], 0);

		//Decide to take another step or wait for further input
		if (dir == 0 && turnOffset == 0)
			stepTracker = NEUTRAL;
		else
			stepTracker = ((stepTracker == GROUP1) ? GROUP2 : GROUP1);
	}
}

void lean(PhidgetRCServoHandle* servos, PhidgetVoltageRatioInputHandle* axes) {

	double yAxis;
	double xAxis;

	PhidgetVoltageRatioInput_getVoltageRatio(axes[0], &xAxis);
	PhidgetVoltageRatioInput_getVoltageRatio(axes[1], &yAxis);

	if (!crouched)
		for (int i = 0; i < NUM_SERVOS; i++)
			PhidgetRCServo_setVelocityLimit(servos[i], isHorizontal(i) ? 180 : 9000);

	setLegPos(servos, LFV, (VMAX / 2) + xAxis * -(VMAX / 4) + yAxis * (VMAX / 4));
	setLegPos(servos, RFV, (VMAX / 2) + xAxis * (VMAX / 4) + yAxis * (VMAX / 4));

	setLegPos(servos, LMV, (VMAX / 2) + xAxis * -(VMAX / 4));
	setLegPos(servos, RMV, (VMAX / 2) + xAxis * (VMAX / 4));

	setLegPos(servos, LRV, (VMAX / 2) + xAxis * -(VMAX / 4) + yAxis * -(VMAX / 4));
	setLegPos(servos, RRV, (VMAX / 2) + xAxis * (VMAX / 4) + yAxis * -(VMAX / 4));
	crouched = 1;
}

void standUp(PhidgetRCServoHandle* servos) {

	if (crouched) {
		for (int i = 0; i < NUM_SERVOS; i++)
			PhidgetRCServo_setVelocityLimit(servos[i], isHorizontal(i) ? 180 : 360);

		setLegPos(servos, LFV, 0);
		setLegPos(servos, RFV, 0);

		setLegPos(servos, LMV, 0);
		setLegPos(servos, RMV, 0);

		setLegPos(servos, LRV, 0);
		setLegPos(servos, RRV, 0);

		crouched = 0;
	}
}

void closeAll(PhidgetRCServoHandle* servos, PhidgetVoltageRatioInputHandle* axes,
	PhidgetDigitalInputHandle button) {
	for (int i = 0; i < NUM_SERVOS; i++) {
		Phidget_close((PhidgetHandle)&servos[i]);
		PhidgetRCServo_delete(&servos[i]);
	}

	for (int i = 0; i < 2; i++) {
		Phidget_close((PhidgetHandle)&axes[i]);
		PhidgetVoltageRatioInput_delete(&axes[i]);
	}

	Phidget_close((PhidgetHandle)&button);
	PhidgetDigitalInput_delete(&button);
}

static void CCONV
onAttachHandler(PhidgetHandle phid, void *ctx) {
	PhidgetReturnCode res;
	int hubPort;
	int channel;
	int serial;
	Phidget_DeviceClass deviceClass;

	res = Phidget_getDeviceSerialNumber(phid, &serial);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to get device serial number\n");
		return;
	}

	res = Phidget_getChannel(phid, &channel);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to get channel number\n");
		return;
	}

	res = Phidget_getHubPort(phid, &hubPort);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to get hub port\n");
		hubPort = -1;
	}

	if (hubPort == -1)
		printf("channel %d on device %d attached\n", channel, serial);
	else
		printf("channel %d on device %d hub port %d attached\n", channel, serial, hubPort);

	Phidget_getDeviceClass(phid, &deviceClass);

	if (deviceClass == PHIDCHCLASS_VOLTAGERATIOINPUT)
		PhidgetVoltageRatioInput_setDataInterval((PhidgetVoltageRatioInputHandle)phid, 20);
}

static void CCONV
onDetachHandler(PhidgetHandle phid, void *ctx) {
	PhidgetReturnCode res;
	int hubPort;
	int channel;
	int serial;

	res = Phidget_getDeviceSerialNumber(phid, &serial);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to get device serial number\n");
		return;
	}

	res = Phidget_getChannel(phid, &channel);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to get channel number\n");
		return;
	}

	res = Phidget_getHubPort(phid, &hubPort);
	if (res != EPHIDGET_OK)
		hubPort = -1;

	if (hubPort != -1)
		printf("channel %d on device %d detached\n", channel, serial);
	else
		printf("channel %d on device %d hub port %d detached\n", channel, hubPort, serial);
}

static void CCONV
errorHandler(PhidgetHandle phid, void *ctx, Phidget_ErrorEventCode errorCode, const char *errorString) {

	fprintf(stderr, "Error: %s (%d)\n", errorString, errorCode);
}

static void CCONV
onPositionChangeHandler(PhidgetRCServoHandle ch, void *ctx, double position) {

	int channel;
	Phidget_getChannel((PhidgetHandle)ch, &channel);

	printf("Position Changed (%d): %.3g\n", channel, position);

}

static void CCONV
onVelocityChangeHandler(PhidgetRCServoHandle ch, void *ctx, double velocity) {

	printf("Velocity Changed: %.3g\n", velocity);
}

static void CCONV
onTargetPositionReachedHandler(PhidgetRCServoHandle ch, void *ctx, double position) {

	int channel;
	Phidget_getChannel((PhidgetHandle)ch, &channel);

	printf("Target Position Reached (%d): %.3g\n", channel, position);

	legsDone[channel] = 1;

}

static void CCONV
onAxisChangeHandler(PhidgetVoltageRatioInputHandle ch, void *ctx, double axisValue) {

}

static void CCONV
onStateChangeHandler(PhidgetDigitalInputHandle ch, void *ctx, int state) {

}

/*
* Creates and initializes the channel.
*/
static PhidgetReturnCode CCONV
initChannel(PhidgetHandle ch) {
	PhidgetReturnCode res;

	res = Phidget_setOnAttachHandler(ch, onAttachHandler, NULL);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to assign on attach handler\n");
		return (res);
	}

	res = Phidget_setOnDetachHandler(ch, onDetachHandler, NULL);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to assign on detach handler\n");
		return (res);
	}

	res = Phidget_setOnErrorHandler(ch, errorHandler, NULL);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to assign on error handler\n");
		return (res);
	}

	return (EPHIDGET_OK);
}

int
main(int argc, char **argv) {
	PhidgetRCServoHandle servos[NUM_SERVOS];
	PhidgetVoltageRatioInputHandle axes[2];
	PhidgetDigitalInputHandle button;
	PhidgetReturnCode res;
	const char *errs;

	int buttonState;

	for (int i = 0; i < NUM_SERVOS; i++) {

		res = PhidgetRCServo_create(&servos[i]);
		Phidget_setIsLocal((PhidgetHandle)&servos[i], 1);
		if (res != EPHIDGET_OK) {
			fprintf(stderr, "failed to create rc servo channel\n");
			exit(1);
		}

		res = initChannel((PhidgetHandle)servos[i]);
		Phidget_setChannel((PhidgetHandle)servos[i], i);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to initialize channel:%s\n", errs);
			exit(1);
		}

		res = PhidgetRCServo_setOnPositionChangeHandler(servos[i], onPositionChangeHandler, NULL);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to set position change handler: %s\n", errs);
			goto done;
		}

		res = PhidgetRCServo_setOnVelocityChangeHandler(servos[i], onVelocityChangeHandler, NULL);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to set velocity change handler: %s\n", errs);
			goto done;
		}

		res = PhidgetRCServo_setOnTargetPositionReachedHandler(servos[i], onTargetPositionReachedHandler, NULL);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to set target position reached handler: %s\n", errs);
			goto done;
		}
	}

	for (int i = 0; i < 2; i++) {

		res = PhidgetVoltageRatioInput_create(&axes[i]);
		if (res != EPHIDGET_OK) {
			fprintf(stderr, "failed to create rc servo channel\n");
			exit(1);
		}

		res = initChannel((PhidgetHandle)axes[i]);
		Phidget_setChannel((PhidgetHandle)axes[i], i);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to initialize channel:%s\n", errs);
			exit(1);
		}

		res = PhidgetVoltageRatioInput_setOnVoltageRatioChangeHandler(axes[i], onAxisChangeHandler, NULL);
		if (res != EPHIDGET_OK) {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to set target position reached handler: %s\n", errs);
			goto done;
		}

	}

	res = PhidgetDigitalInput_create(&button);
	if (res != EPHIDGET_OK) {
		fprintf(stderr, "failed to create rc servo channel\n");
		exit(1);
	}

	/*
	* Open the channel synchronously: waiting a maximum of 5 seconds.
	*/
	for (int i = 0; i < NUM_SERVOS; i++) {
		res = Phidget_openWaitForAttachment((PhidgetHandle)servos[i], 5000);
		if (res != EPHIDGET_OK) {
			if (res == EPHIDGET_TIMEOUT) {
				printf("Channel did not attach after 5 seconds: please check that the device is attached\n");
			} else {
				Phidget_getErrorDescription(res, &errs);
				fprintf(stderr, "failed to open channel:%s\n", errs);
			}
			goto done;

		}
	}
	/*
	* Open the channel synchronously: waiting a maximum of 5 seconds.
	*/
	for (int i = 0; i < 2; i++) {
		res = Phidget_openWaitForAttachment((PhidgetHandle)axes[i], 5000);
		if (res != EPHIDGET_OK) {
			if (res == EPHIDGET_TIMEOUT) {
				printf("Channel did not attach after 5 seconds: please check that the device is attached\n");
			} else {
				Phidget_getErrorDescription(res, &errs);
				fprintf(stderr, "failed to open channel:%s\n", errs);
			}
			goto done;
		}
	}

	res = initChannel((PhidgetHandle)button);
	if (res != EPHIDGET_OK) {
		Phidget_getErrorDescription(res, &errs);
		fprintf(stderr, "failed to initialize channel:%s\n", errs);
		exit(1);
	}

	res = PhidgetDigitalInput_setOnStateChangeHandler(button, onStateChangeHandler, NULL);
	if (res != EPHIDGET_OK) {
		Phidget_getErrorDescription(res, &errs);
		fprintf(stderr, "failed to set target position reached handler: %s\n", errs);
		goto done;
	}

	/*
	* Open the channel synchronously: waiting a maximum of 5 seconds.
	*/
	res = Phidget_openWaitForAttachment((PhidgetHandle)button, 5000);
	if (res != EPHIDGET_OK) {
		if (res == EPHIDGET_TIMEOUT) {
			printf("Channel did not attach after 5 seconds: please check that the device is attached\n");
		} else {
			Phidget_getErrorDescription(res, &errs);
			fprintf(stderr, "failed to open channel:%s\n", errs);
		}
		goto done;
	}

	ssleep(1);

	for (int i = 0; i < NUM_SERVOS; i++) {
		if (isHorizontal(i))
			setLegPos(servos, i, 0);
		else
			setLegPos(servos, i, 0);

		double maxVel;
		double maxAcc;

		PhidgetRCServo_getMaxVelocityLimit(servos[i], &maxVel);
		PhidgetRCServo_getMaxAcceleration(servos[i], &maxAcc);

		PhidgetRCServo_setVelocityLimit(servos[i], isHorizontal(i) ? 180 : 360);

		PhidgetRCServo_setAcceleration(servos[i], isHorizontal(i) ? 360 : maxAcc);

		printf("Setting engaged\n");
		res = PhidgetRCServo_setEngaged(servos[i], 1);
		if (res != EPHIDGET_OK) {
			printf("failed to set engaged\n");
		}
	}

	ssleep(1);
	while (1) {
		PhidgetDigitalInput_getState(button, &buttonState);
		if (!buttonState) {
			if (crouched)
				standUp(servos);
			walk(servos, axes);
		} else
			lean(servos, axes);

		ssleep(0.02);
	}

done:
	closeAll(servos, axes, button);

	exit(res);
}

static void CCONV
ssleep(double tm) {
#ifdef _WIN32
	Sleep((int)(tm * 1000));
#else
	usleep(tm * 1000000);
#endif
}

